package com.ociotec.tmapi.graphql;

import io.smallrye.graphql.client.core.Argument;
import io.smallrye.graphql.client.core.InputObjectField;
import static io.smallrye.graphql.client.core.Argument.arg;
import static io.smallrye.graphql.client.core.Document.document;
import static io.smallrye.graphql.client.core.Field.field;
import static io.smallrye.graphql.client.core.Enum.gqlEnum;
import static io.smallrye.graphql.client.core.InputObject.inputObject;
import static io.smallrye.graphql.client.core.InputObjectField.prop;
import static io.smallrye.graphql.client.core.Operation.operation;

import java.io.InvalidClassException;
import java.time.Instant;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.atomic.AtomicReference;
import java.util.stream.Collectors;

import org.eclipse.microprofile.graphql.GraphQLApi;
import org.eclipse.microprofile.graphql.Query;

import com.ociotec.tmapi.graphql.models.Property;
import com.ociotec.tmapi.graphql.models.PropertyFilter;
import com.ociotec.tmapi.graphql.models.Sample;
import com.ociotec.tmapi.graphql.models.Value;
import com.ociotec.tmapi.graphql.models.ValueInt;
import com.ociotec.tmapi.graphql.models.ValueLong;
import com.ociotec.tmapi.graphql.models.ValueString;
import com.ociotec.tmapi.graphql.models.ValueTimestamp;

import graphql.language.Field;
import graphql.language.SelectionSet;

import com.ociotec.tmapi.graphql.models.ValueBoolean;
import com.ociotec.tmapi.graphql.models.ValueFloat;
import com.ociotec.tmapi.graphql.models.ValueDouble;

import io.smallrye.graphql.client.GraphQLClient;
import io.smallrye.graphql.client.core.Document;
import io.smallrye.graphql.client.core.FieldOrFragment;
import io.smallrye.graphql.client.core.OperationType;
import io.smallrye.graphql.client.dynamic.api.DynamicGraphQLClient;
import io.smallrye.graphql.execution.context.SmallRyeContext;
import io.micrometer.core.instrument.Counter;
import io.micrometer.core.instrument.MeterRegistry;
import io.quarkus.logging.Log;
import io.quarkus.vertx.ConsumeEvent;
import io.smallrye.graphql.api.Subscription;
import io.smallrye.mutiny.Multi;
import io.smallrye.mutiny.operators.multi.processors.BroadcastProcessor;
import io.vertx.core.cli.annotations.Description;
import io.vertx.mutiny.core.eventbus.EventBus;
import jakarta.enterprise.context.ApplicationScoped;
import jakarta.inject.Inject;
import jakarta.json.JsonNumber;
import jakarta.json.JsonObject;
import jakarta.json.JsonString;
import jakarta.json.JsonValue;

@GraphQLApi
@ApplicationScoped
public class SamplesGraphqlResource {

    @Inject
    @GraphQLClient("db-api")
    DynamicGraphQLClient graphQLClient;

    @Inject
    SmallRyeContext context;

    @Inject
    MeterRegistry meterRegistry;

    @Query
    @Description("Returns stored telemetry filtered by timestamp, system, subsystem and/or name")
    public List<Sample> getSamples(
        String system,
        String subsystem,
        String name,
        Instant begin,
        Instant end,
        Integer limit,
        Integer offset,
        Map<String, PropertyFilter> properties
    ) throws ExecutionException, InterruptedException, InvalidClassException {
        Counter.builder("graphql")
            .baseUnit("operations")
            .description("TM API samples queries")
            .tag("operation_type", "query")
            .tag("operation_name", "samples")
            .tag("counter_type", "total")
            .register(meterRegistry)
            .increment();

        var arguments = createArguments(system, subsystem, name, begin, end, limit, offset, properties);
        var requestedFields = extractRequestedFieldsFromContext();
        var fields = generateFieldsToRequest(requestedFields);
        // https://stackoverflow.com/questions/68772614/quarkus-smallrye-graphql-client-mutation-query
        Document query = document(
            operation(
                OperationType.QUERY,
                field("samples", arguments, fields)
            )
        );
        var response = graphQLClient.executeSync(query);

        List<Sample> samples = null;
        var data = response.getData();
        if (data != null) {
            var jsonSamples = data.getJsonArray("samples");
            if (jsonSamples != null) {
                samples = new ArrayList<Sample>(jsonSamples.size());
                var jsonIterator = jsonSamples.iterator();
                while (jsonIterator.hasNext()) {
                    var jsonSample = jsonIterator.next().asJsonObject();
                    samples.add(createSample(requestedFields, jsonSample));
                }

                Counter.builder("graphql")
                .baseUnit("samples")
                .description("TM samples served by TM API on samples queries")
                .tag("operation_type", "retrieval")
                .register(meterRegistry)
                .increment(samples.size());

                Counter.builder("graphql")
                .baseUnit("operations")
                .description("TM API samples queries")
                .tag("operation_type", "query")
                .tag("operation_name", "samples")
                .tag("counter_type", "success")
                .register(meterRegistry)
                .increment();
            }
        }
        
        return samples;
    }

    AtomicReference<Sample> lastLiveSample = new AtomicReference<>();

    @Query
    @Description("Returns last live telemetry sample")
    public Sample lastLiveSample() {
        Counter.builder("graphql")
            .baseUnit("operations")
            .description("TM API samples queries")
            .tag("operation_type", "query")
            .tag("operation_name", "lastLiveSample")
            .tag("counter_type", "total")
            .register(meterRegistry)
            .increment();
        Counter.builder("graphql")
            .baseUnit("samples")
            .description("TM samples served by TM API on samples queries")
            .tag("operation_type", "live")
            .register(meterRegistry)
            .increment(lastLiveSample == null ? 0 : 1);
        Counter.builder("graphql")
            .baseUnit("operations")
            .description("TM API samples queries")
            .tag("operation_type", "query")
            .tag("operation_name", "lastLiveSample")
            .tag("counter_type", "success")
            .register(meterRegistry)
            .increment();
        return lastLiveSample.get();
    }

    BroadcastProcessor<List<Sample>> samplesProcessor = BroadcastProcessor.create();

    @Inject
    EventBus newSystemBus;

    @Subscription
    @Description("Subscription for new samples, it can be optionally filtered by system, subsystem and/or name")
    public Multi<List<Sample>> newSamples(String system, String subsystem, String name) {
        Counter.builder("graphql")
            .baseUnit("operations")
            .description("TM API samples queries")
            .tag("operation_type", "subscription")
            .tag("operation_name", "newSamples")
            .tag("counter_type", "total")
            .register(meterRegistry)
            .increment();
        if (system != null) {
            newSystemBus.publish("add-system", system);
        }
        var result = samplesProcessor.map(samples ->
            samples.stream().filter(sample ->
                (system == null || system.equals(sample.getSystem())) &&
                (subsystem == null || subsystem.equals(sample.getSubsystem())) &&
                (name == null || name.equals(sample.getName()))
            ).collect(Collectors.toList())
        ).filter(samples -> !samples.isEmpty())
        .onCancellation().invoke(() -> {
            newSystemBus.publish("remove-system", system);
        });
        result.subscribe().with(
            samples -> {
                Counter.builder("graphql")
                    .baseUnit("samples")
                    .description("TM samples served by TM API on samples queries")
                    .tag("operation_type", "live")
                    .register(meterRegistry)
                    .increment(samples.size());
            }
        );
        Counter.builder("graphql")
            .baseUnit("operations")
            .description("TM API samples queries")
            .tag("operation_type", "subscription")
            .tag("operation_name", "newSamples")
            .tag("counter_type", "success")
            .register(meterRegistry)
            .increment();
        return result;
    }

    @ConsumeEvent("samples")
    public void consumeSample(List<Sample> samples) {
        samplesProcessor.onNext(samples);
        var systems = new HashSet<String>();
        for (var sample : samples) {
            systems.add(sample.getSystem());
        }
        Log.info("Processed x" + samples.size() + " TM samples from systems: " + systems);
        lastLiveSample.set(samples.getLast());
    }

    private Set<String> extractRequestedFieldsFromContext() {
        // https://stackoverflow.com/a/78032435
        var requestedFields = new HashSet<String>();
        var children = context.getDataFetchingEnvironment().getField().getChildren();
        for (var child : children) {
            if (child instanceof SelectionSet) {
                var nodes = ((SelectionSet) child).getChildren();
                for (var node : nodes) {
                    if (node instanceof Field) {
                        requestedFields.add(((Field) node).getName());
                    }
                }
            }
        }
        return requestedFields;
    }

    private static final int MAX_QUERY_LIMIT = 1000;
    private static final int DEFAULT_QUERY_LIMIT = 100;
    private static final int DEFAULT_QUERY_OFFSET = 0;
    private List<Argument> createArguments(
        String system,
        String subsystem,
        String name,
        Instant begin,
        Instant end,
        Integer limit,
        Integer offset,
        Map<String, PropertyFilter> properties
    ) {
        var arguments = new ArrayList<Argument>();
        int limitToUse = DEFAULT_QUERY_LIMIT;
        if (limit != null) {
            if (limit >= 1 && limit <= MAX_QUERY_LIMIT) {
                limitToUse = limit;
            }
        }
        int offsetToUse = DEFAULT_QUERY_OFFSET;
        if (offset != null) {
            if (offset >= 0) {
                offsetToUse = offset;
            }
        }
        arguments.add(arg("limit", limitToUse));
        arguments.add(arg("offset", offsetToUse));
        arguments.add(arg("order_by", inputObject(
            prop("timestamp", gqlEnum("asc")),
            prop("system", gqlEnum("asc")),
            prop("subsystem", gqlEnum("asc")),
            prop("name", gqlEnum("asc"))
        )));
        if (system != null || subsystem != null || name != null || begin != null || end != null || properties != null) {
            var equals = new ArrayList<InputObjectField>();
            if (system != null) {
                equals.add(prop("system", inputObject(prop("_eq", system))));
            }
            if (subsystem != null) {
                equals.add(prop("subsystem", inputObject(prop("_eq", subsystem))));
            }
            if (name != null) {
                equals.add(prop("name", inputObject(prop("_eq", name))));
            }
            if (begin != null || end != null) {
                var timestampProps = new ArrayList<InputObjectField>();
                if (begin != null) {
                    timestampProps.add(prop("_gte", begin.toString()));
                }
                if (end != null) {
                    timestampProps.add(prop("_lte", end.toString()));
                }
                equals.add(prop("timestamp", inputObject(timestampProps.toArray(new InputObjectField[0]))));
            }
            if (properties != null) {
                var propertiesFilter = new ArrayList<InputObjectField>();
                for (var property : properties.entrySet()) {
                    if (property.getValue().getIntValue() != null) {
                        propertiesFilter.add(prop(property.getKey(), property.getValue().getIntValue()));
                    } else if (property.getValue().getLongValue() != null) {
                        propertiesFilter.add(prop(property.getKey(), property.getValue().getLongValue()));
                    } else if (property.getValue().getStringValue() != null) {
                        propertiesFilter.add(prop(property.getKey(), property.getValue().getStringValue()));
                    } else if (property.getValue().getBooleanValue() != null) {
                        propertiesFilter.add(prop(property.getKey(), property.getValue().getBooleanValue()));
                    } else if (property.getValue().getFloatValue() != null) {
                        propertiesFilter.add(prop(property.getKey(), property.getValue().getFloatValue()));
                    } else if (property.getValue().getDoubleValue() != null) {
                        propertiesFilter.add(prop(property.getKey(), property.getValue().getDoubleValue()));
                    }
                }
                equals.add(prop("properties", inputObject(prop("_contains", inputObject(propertiesFilter.toArray(new InputObjectField[0]))))));
            }
            arguments.add(arg("where", inputObject(equals.toArray(new InputObjectField[0]))));
        }
        return arguments;
    }

    private FieldOrFragment[] generateFieldsToRequest(Set<String> requestedFields) {
        var fields = new ArrayList<FieldOrFragment>();
        if (requestedFields.contains("id")) {
            fields.add(field("id"));
        }
        if (requestedFields.contains("system")) {
            fields.add(field("system"));
        }
        if (requestedFields.contains("subsystem")) {
            fields.add(field("subsystem"));
        }
        if (requestedFields.contains("name")) {
            fields.add(field("name"));
        }
        if (requestedFields.contains("timestamp")) {
            fields.add(field("timestamp"));
        }
        if (requestedFields.contains("value")) {
            fields.add(field("valueInt"));
            fields.add(field("valueLong"));
            fields.add(field("valueString"));
            fields.add(field("valueBoolean"));
            fields.add(field("valueFloat"));
            fields.add(field("valueDouble"));
            fields.add(field("valueTimestamp"));
        }
        if (requestedFields.contains("properties")) {
            fields.add(field("properties"));
        }
        return fields.toArray(new FieldOrFragment[0]);
    }

    private Value createValue(JsonObject jsonSample) {
        Value value = null;
        if (!jsonSample.isNull("valueInt")) {
            value = ValueInt.builder()
            .value(jsonSample.getJsonNumber("valueInt").numberValue().intValue())
            .build();
        } else if (!jsonSample.isNull("valueLong")) {
            value = ValueLong.builder()
                .value(jsonSample.getJsonNumber("valueLong").numberValue().longValue())
                .build();
        } else if (!jsonSample.isNull("valueString")) {
            value = ValueString.builder()
                .value(jsonSample.getString("valueString"))
                .build();
        } else if (!jsonSample.isNull("valueBoolean")) {
            value = ValueBoolean.builder()
                .value(jsonSample.getBoolean("valueBoolean"))
                .build();
        } else if (!jsonSample.isNull("valueFloat")) {
            value = ValueFloat.builder()
                .value(jsonSample.getJsonNumber("valueFloat").numberValue().floatValue())
                .build();
        } else if (!jsonSample.isNull("valueDouble")) {
            value = ValueDouble.builder()
                .value(jsonSample.getJsonNumber("valueDouble").numberValue().doubleValue())
                .build();
        } else if (!jsonSample.isNull("valueTimestamp")) {
            value = ValueTimestamp.builder()
                .value(Instant.parse(jsonSample.getString("valueTimestamp") + "Z"))
                .build();
        } else {
            Log.error("Invalid value class for sample: " + jsonSample);
        }
        return value;
    }

    private Sample createSample(Set<String> requestedFields, JsonObject jsonSample) {
        var sampleBuilder = Sample.builder();
        if (requestedFields.contains("id") && !jsonSample.isNull("id")) {
            sampleBuilder.id(UUID.fromString(jsonSample.getString("id")));
        }
        if (requestedFields.contains("system") && !jsonSample.isNull("system")) {
            sampleBuilder.system(jsonSample.getString("system"));
        }
        if (requestedFields.contains("subsystem") && !jsonSample.isNull("subsystem")) {
            sampleBuilder.subsystem(jsonSample.getString("subsystem"));
        }
        if (requestedFields.contains("name") && !jsonSample.isNull("name")) {
            sampleBuilder.name(jsonSample.getString("name"));
        }
        if (requestedFields.contains("timestamp") && !jsonSample.isNull("timestamp")) {
            sampleBuilder.timestamp(Instant.parse(jsonSample.getString("timestamp") + "Z"));
        }
        if (requestedFields.contains("value")) {
            var value = createValue(jsonSample);
            if (value != null) {
                sampleBuilder.value(value);
            }
        }
        if (requestedFields.contains("properties") && !jsonSample.isNull("properties")) {
            var properties = new ArrayList<Property>();
            var jsonProperties = jsonSample.getJsonObject("properties");
            for (var property : jsonProperties.entrySet()) {
                var propertyKey = property.getKey();
                var propertyValue = createProperty(property.getValue());
                properties.add(
                    Property.builder()
                        .key(propertyKey)
                        .value(propertyValue)
                        .build()
                );
            }
            sampleBuilder.properties(properties);
        }
        return sampleBuilder.build();
    }

    private Value createProperty(JsonValue anyValue) {
        Value propertyValue = null;
        if (anyValue instanceof JsonString) {
            propertyValue = ValueString.builder()
                .value(((JsonString) anyValue).getString())
                .build();
        } else if (anyValue instanceof JsonNumber) {
            var numberValue = (JsonNumber) anyValue;
            if (numberValue.isIntegral()) {
                propertyValue = ValueLong.builder()
                    .value(((JsonNumber) anyValue).longValue())
                    .build();
            } else {
                propertyValue = ValueDouble.builder()
                    .value(numberValue.doubleValue())
                    .build();
            }
        } else if (anyValue.equals(JsonValue.TRUE)) {
            propertyValue = ValueBoolean.builder()
                .value(true)
                .build();
        } else if (anyValue.equals(JsonValue.FALSE)) {
            propertyValue = ValueBoolean.builder()
                .value(false)
                .build();
        }
        return propertyValue;
    }

}
