package com.ociotec.tmapi.kafka;

import java.io.InvalidClassException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.apache.kafka.common.TopicPartition;
import org.eclipse.microprofile.reactive.messaging.Incoming;

import com.ociotec.models.Sample;
import com.ociotec.models.Timestamp;
import com.ociotec.tmapi.graphql.models.Property;

import io.quarkus.logging.Log;
import io.quarkus.vertx.ConsumeEvent;
import io.smallrye.reactive.messaging.kafka.KafkaClientService;
import io.vertx.mutiny.core.eventbus.EventBus;
import jakarta.enterprise.context.ApplicationScoped;
import jakarta.inject.Inject;

@ApplicationScoped
public class SamplesKafkaConsumer {
    
    @Inject
    EventBus samplesBus;

    @Incoming("samples")
    public void consume(List<Sample> samples) throws InvalidClassException {
        var apiSamples = new ArrayList<com.ociotec.tmapi.graphql.models.Sample>(samples.size());
        for (var sample : samples) {
            Log.debug("New sample consumed from Kafka: " + sample);
            var value = sample.getValue();
            var apiValue = convertValue(value, "value");
            var sampleProperties = new ArrayList<Property>();
            for (var property : sample.getProperties().entrySet()) {
                var apiPropertyValue = convertValue(property.getValue(), "property value");
                var apiProperty = Property.builder()
                    .key(property.getKey())
                    .value(apiPropertyValue)
                    .build();
                sampleProperties.add(apiProperty);
            }
            var apiSample = com.ociotec.tmapi.graphql.models.Sample.builder()
                .id(sample.getId())
                .system(sample.getSystem())
                .subsystem(sample.getSubsystem())
                .name(sample.getName())
                .timestamp(sample.getTimestamp())
                .value(apiValue)
                .properties(sampleProperties)
                .build();
            apiSamples.add(apiSample);
        }
        samplesBus.publish("samples", apiSamples);
    }

    private com.ociotec.tmapi.graphql.models.Value convertValue(Object value, String kindOfValue) throws InvalidClassException {
        com.ociotec.tmapi.graphql.models.Value apiValue = null;
        if (value instanceof Integer) {
            apiValue = com.ociotec.tmapi.graphql.models.ValueInt.builder().value((Integer) value).build();
        } else if (value instanceof Long) {
            apiValue = com.ociotec.tmapi.graphql.models.ValueLong.builder().value((Long) value).build();
        } else if (value instanceof String) {
            apiValue = com.ociotec.tmapi.graphql.models.ValueString.builder().value((String) value).build();
        } else if (value instanceof Boolean) {
            apiValue = com.ociotec.tmapi.graphql.models.ValueBoolean.builder().value((Boolean) value).build();
        } else if (value instanceof Float) {
            apiValue = com.ociotec.tmapi.graphql.models.ValueFloat.builder().value((Float) value).build();
        } else if (value instanceof Double) {
            apiValue = com.ociotec.tmapi.graphql.models.ValueDouble.builder().value((Double) value).build();
        } else if (value instanceof Timestamp) {
            apiValue = com.ociotec.tmapi.graphql.models.ValueTimestamp.builder().value(((Timestamp) value).getTimestamp()).build();
        } else  {
            throw new InvalidClassException("Invalid " + kindOfValue + " class: " + value.getClass().getCanonicalName());
        }
        return apiValue;
    }

    private static String[] SYSTEMS_ARRAY = {"SAT01", "SAT02", "SAT03", "SAT04", "SAT05", "SAT06", "SAT07", "SAT08", "SAT09", "SAT10"};
    private static List<String> SYSTEMS = Arrays.asList(SYSTEMS_ARRAY);

    private int partitionForSystem(String system) {
        return SYSTEMS.indexOf(system);
    }

    @Inject
    KafkaClientService kafka;

    @ConsumeEvent("add-system")
    public void addSystem(String system) {
        var partition = partitionForSystem(system);
        Log.info("Allowing samples from system: " + system + " on partition: " + partition);
        var consumer = kafka.getConsumer("samples");
        var partitions = new ArrayList<TopicPartition>();
        partitions.add(new TopicPartition("samples", partition));
        consumer.seekToEnd(partitions);
    }

    @ConsumeEvent("remove-system")
    public void removeSystem(String system) {
        var partition = partitionForSystem(system);
        Log.info("Removing samples from system: " + system + " on partition: " + partition);
    }

}
